#ifndef CURVED_FUNCTION_HH
#define CURVED_FUNCTION_HH

#include <dune/fufem/interval.hh>

namespace Dune {
template <class NonlinearityType> class CurvedFunction {
  typedef typename NonlinearityType::VectorType VectorType;
  typedef typename NonlinearityType::MatrixType MatrixType;

public:
  CurvedFunction(MatrixType const &A, VectorType const &b,
                 NonlinearityType const &phi, VectorType const &x,
                 VectorType const &dir)
      : A(A), b(b), phi(phi), x(x), dir(dir) {}

  double quadraticPart() const { return 0; }

  double linearPart() const { return 0; }

  void subDiff(double m, Interval<double> &D) const {
    VectorType x;
    evaluate(m, x);
    VectorType tangent;
    tangentialDirection(m, tangent);

    phi.directionalSubDiff(x, tangent, D);

    VectorType tmp;
    A.mv(x, tmp);                      //  Ax
    tmp -= b;                          //  Ax - b
    double const dotp = tmp * tangent; // <Ax - b,t>
    D[0] += dotp;
    D[1] += dotp;
  }

  void domain(Interval<double> &domain) const {
    // TODO
    domain[0] = 0;
    domain[1] = 1;
  }

  void evaluate(double m, VectorType &y) const {
    y = 0;
    y.axpy(1 - m, x);
    y.axpy(m, dir);
    y /= y.two_norm();
  }

private:
  MatrixType const &A;
  VectorType const &b;
  NonlinearityType const &phi;
  VectorType const &x;
  VectorType const &dir;

  /*
    Tangential vector within the plane, with positive direction.

    < (1-m)x + m*y,-m*|y|^2*x+(1-m)*|x|^2*y>
    = -m(1-m)|y|^2<x,x> + m(1-m)|x|^2<y,y> (because <x,y> = 0)
    = 0
  */
  void tangentialDirection(double m, VectorType &y) const {
    y = 0;
    y.axpy(-m * dir.two_norm2(), x);
    y.axpy((1 - m) * x.two_norm2(), dir);
  }
};
}

#endif
