#ifndef DUNE_TECTONIC_GLOBAL_LAURSEN_NONLINEARITY_HH
#define DUNE_TECTONIC_GLOBAL_LAURSEN_NONLINEARITY_HH

#include <vector>

#include <dune/common/fvector.hh>
#include <dune/common/shared_ptr.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/bvector.hh>

#include "globalnonlinearity.hh"
#include "localnonlinearity.hh"
#include "nicefunction.hh"

namespace Dune {
template <class OuterFunctionType, class VectorType, class MatrixType>
class GlobalLaursenNonlinearity
    : public GlobalNonlinearity<VectorType, MatrixType> {
private:
  typedef shared_ptr<BlockVector<FieldVector<double, 1>> const> dataptr;

public:
  using GlobalNonlinearity<VectorType, MatrixType>::dim;

  GlobalLaursenNonlinearity(dataptr mu, dataptr normalStress,
                            dataptr nodalIntegrals)
      : mu(mu),
        normalStress(normalStress),
        nodalIntegrals(nodalIntegrals),
        trivialNonlinearity(
            new LocalNonlinearity<dim>(make_shared<TrivialFunction const>())),
        restrictions(nodalIntegrals->size()) // TODO: can we get the size from
                                             // another place?
  {
    for (auto &x : restrictions)
      x = shared_ptr<LocalNonlinearity<dim> const>();
  }

  /*
    Return a restriction of the outer function to the i'th node. If
    mu and sigma_n denote the coefficient of friction and the normal
    stress, respectively, at the i'th node, this function is given
    by

    sigma_n [(1/eta \bar Gamma)* + mu id]

    TODO: We currently assume eta = 1
  */
  virtual shared_ptr<LocalNonlinearity<dim> const> restriction(int i) const {
    if ((*nodalIntegrals)[i] == 0)
      return trivialNonlinearity;

    if (restrictions[i] != nullptr)
      return restrictions[i];

    double coefficient = (*nodalIntegrals)[i];
    coefficient *= (*normalStress)[i];
    coefficient *= 1 + (*mu)[i];

    auto const func = make_shared<OuterFunctionType const>(coefficient);
    restrictions[i] = make_shared<LocalNonlinearity<dim> const>(func);
    return restrictions[i];
  }

private:
  dataptr mu;
  dataptr normalStress;
  dataptr nodalIntegrals;

  shared_ptr<LocalNonlinearity<dim> const> const trivialNonlinearity;

  std::vector<shared_ptr<LocalNonlinearity<dim> const>> mutable restrictions;
};
}
#endif
